#ifndef _GRAPHICS_H
#define _GRAPHICS_H

#include <string>
#include <ostream>
#include <map>
#include <list>
#include <vector>

#include "shader.h"
//#include "model_ase.h"
#include "mathvector.h"
#include "fbtexture.h"
#include "scenegraph.h"
#include "matrix4.h"
#include "texture.h"
#include "reseatable_reference.h"

#include <SDL/SDL.h>

class SCENENODE;

class GRAPHICS_SDLGL
{
private:
	class GLSTATEMANAGER
	{
		private:
			std::vector <bool> used; //on modern compilers this should result in a lower memory usage bit_vector-type arrangement
			std::vector <bool> state;
			
			void Set(int stateid, bool newval)
			{
				assert(stateid <= 65535);
				
				if (used[stateid])
				{
					if (state[stateid] != newval)
					{
						state[stateid] = newval;
						if (newval)
							glEnable(stateid);
						else
							glDisable(stateid);
					}
				}
				else
				{
					used[stateid] = true;
					state[stateid] = newval;
					if (newval)
						glEnable(stateid);
					else
						glDisable(stateid);
				}
			}
			
		public:
			GLSTATEMANAGER() : used(65536, false), state(65536)
			{
			}
			
			inline void Enable(int stateid)
			{
				Set(stateid, true);
			}
			
			inline void Disable(int stateid)
			{
				Set(stateid, false);
			}
	} glstate;
	
	///purely abstract base class
	class RENDER_INPUT
	{
		public:
			virtual void Render(GLSTATEMANAGER & glstate) = 0; ///< must be implemented by derived classes
	};
	
	class RENDER_INPUT_POSTPROCESS : public RENDER_INPUT
	{
		private:
			const FBTEXTURE_GL * source_texture;
			SHADER_GLSL * shader;
			
		public:
			RENDER_INPUT_POSTPROCESS() : source_texture(NULL), shader(NULL) {}
			
			void SetSourceTexture(FBTEXTURE_GL & newsource)
			{
				source_texture = &newsource;
			}
			
			void SetShader(SHADER_GLSL * newshader) {shader = newshader;}
			
			virtual void Render(GLSTATEMANAGER & glstate)
			{
				assert(source_texture);
				assert(shader);
				
				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
				
				shader->Enable();
				
				glMatrixMode(GL_PROJECTION);
				glPushMatrix();
				glLoadIdentity();
				glOrtho( 0, 1, 0, 1, -1, 1 );
				glMatrixMode(GL_MODELVIEW);
				glPushMatrix();
				glLoadIdentity();
	
				glColor4f(1,1,1,1);
				glstate.Disable(GL_BLEND);
				glstate.Disable(GL_DEPTH_TEST);
				glstate.Enable(GL_TEXTURE_2D);
				
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_NONE);
				glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_LUMINANCE);
				source_texture->Activate();

				glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.0f,  0.0f,  0.0f);
				glTexCoord2f(1.0, 0.0f); glVertex3f( 1.0f,  0.0f,  0.0f);
				glTexCoord2f(1.0, 1.0); glVertex3f( 1.0f,  1.0f,  0.0f);
				glTexCoord2f(0.0f, 1.0); glVertex3f( 0.0f,  1.0f,  0.0f);
				glEnd();
	
				glMatrixMode(GL_PROJECTION);
				glPopMatrix();
				glMatrixMode(GL_MODELVIEW);
				glPopMatrix();

				glstate.Enable(GL_DEPTH_TEST);
				glstate.Disable(GL_TEXTURE_2D);
			}
	};
	
	class RENDER_INPUT_SCENE : public RENDER_INPUT
	{
		public:
			enum SHADER_TYPE
			{
				SHADER_SIMPLE,
    				SHADER_DISTANCEFIELD,
				SHADER_FULL,
				SHADER_FULLBLEND,
    				SHADER_SKYBOX,
				SHADER_NONE
			};
		
		private:
			std::vector <SCENEDRAW> * drawlist;
			bool last_transform_valid;
			MATRIX4 <float> last_transform;
			QUATERNION <float> cam_rotation; //used for the skybox effect
			MATHVECTOR <float, 3> cam_position;
			MATHVECTOR <float, 3> lightposition;
			MATHVECTOR <float, 3> orthomin;
			MATHVECTOR <float, 3> orthomax;
			float w, h;
			float camfov;
			float frustum[6][4]; //used for frustum culling
			float lod_far; //used for distance culling
			bool shaders;
			bool clearcolor, cleardepth;
			void SetActiveShader(const SHADER_TYPE & newshader);
			std::vector <SHADER_GLSL *> shadermap;
			SHADER_TYPE activeshader;
			reseatable_reference <TEXTURE_GL> reflection;
			bool orthomode;
			unsigned int fsaa;
			
			void DrawList(std::vector <SCENEDRAW> & drawlist, GLSTATEMANAGER & glstate);
			bool FrustumCull(SCENEDRAW & tocull) const;
			void SelectAppropriateShader(SCENEDRAW & forme);
			void SelectFlags(SCENEDRAW & forme, GLSTATEMANAGER & glstate);
			void SelectTexturing(SCENEDRAW & forme, GLSTATEMANAGER & glstate);
			bool SelectTransformStart(SCENEDRAW & forme, GLSTATEMANAGER & glstate);
			void SelectTransformEnd(SCENEDRAW & forme, bool need_pop);
			void ExtractFrustum();
			
		public:
			RENDER_INPUT_SCENE() : drawlist(NULL), last_transform_valid(false), shaders(false), clearcolor(false), cleardepth(false), orthomode(false)
			{
				shadermap.resize(SHADER_NONE, NULL);
				MATHVECTOR <float, 3> front(1,0,0);
				lightposition = front;
				QUATERNION <float> ldir;
				ldir.Rotate(3.141593*0.4,0,0,1);
				ldir.Rotate(3.141593*0.5*0.7,1,0,0);
				ldir.RotateVector(lightposition);
			}
			
			void SetDrawList(std::vector <SCENEDRAW> * ndrawlist) {drawlist = ndrawlist;}
			void DisableOrtho() {orthomode = false;}
			void SetOrtho(const MATHVECTOR <float, 3> & neworthomin, const MATHVECTOR <float, 3> & neworthomax) {orthomode = true; orthomin = neworthomin; orthomax = neworthomax;}
			void SetCameraInfo(const MATHVECTOR <float, 3> & newpos, const QUATERNION <float> & newrot, float newfov, float newlodfar, float neww, float newh);
			void SetSunDirection(const MATHVECTOR <float, 3> & newsun) {lightposition = newsun;}
			void SetFlags(bool newshaders) {shaders=newshaders;}
			void SetDefaultShader(SHADER_GLSL & newdefault) {shadermap.clear();shadermap.resize(SHADER_NONE, &newdefault);}
			void SetShader(SHADER_TYPE stype, SHADER_GLSL & newshader) {assert((unsigned int)stype < shadermap.size());shadermap[stype]=&newshader;}
			void SetClear(bool newclearcolor, bool newcleardepth) {clearcolor = newclearcolor;cleardepth = newcleardepth;}
			virtual void Render(GLSTATEMANAGER & glstate);
			void SetReflection ( TEXTURE_GL & value ) {reflection = value;}

	void SetFSAA ( unsigned int value )
	{
		fsaa = value;
	}
	
	};
	
	class RENDER_OUTPUT
	{
		private:
			FBTEXTURE_GL fbo;
			enum
			{
				RENDER_TO_FBO,
    				RENDER_TO_FRAMEBUFFER
			} target;
		
		public:
			RENDER_OUTPUT() : target(RENDER_TO_FRAMEBUFFER) {}
			
			///returns the FBO that the user should set up as necessary
			FBTEXTURE_GL & RenderToFBO()
			{
				target = RENDER_TO_FBO;
				return fbo;
			}
			
			void RenderToFramebuffer()
			{
				target = RENDER_TO_FRAMEBUFFER;
			}
			
			void Begin(std::ostream & error_output)
			{
				if (target == RENDER_TO_FBO)
					fbo.Begin(error_output);
			}
			
			void End(std::ostream & error_output)
			{
				if (target == RENDER_TO_FBO)
					fbo.End(error_output);
			}
	};
	
	int w, h;
	SDL_Surface * surface;
	bool initialized;
	bool using_shaders;
	GLint max_anisotropy;
	bool shadows;
	int shadow_distance;
	int shadow_quality;
	enum {REFLECTION_DISABLED, REFLECTION_STATIC, REFLECTION_DYNAMIC} reflection_status;
	TEXTURE_GL static_reflection;
	
	std::map <std::string, SHADER_GLSL> shadermap;
	std::map <std::string, SHADER_GLSL>::iterator activeshader;
	
	std::map < DRAWABLE_FILTER *, std::vector <SCENEDRAW> > drawlist_map1; //this is cached between frames to avoid having to reallocate large blocks of memory every frame
	std::map < DRAWABLE_FILTER *, std::vector <SCENEDRAW> > drawlist_map2; //this is double buffered to allow fast multithreaded syncing
	bool drawlist_map1_is_front; // true if the map1 drawlist is currently being rendered from
	std::list <DRAWABLE_FILTER *> filter_list;
	DRAWABLE_FILTER no2d_noblend;
	DRAWABLE_FILTER no2d_blend;
	DRAWABLE_FILTER only2d;
	DRAWABLE_FILTER skyboxes_noblend;
	DRAWABLE_FILTER skyboxes_blend;
	DRAWABLE_FILTER camtransfilter;
	
	//render pipeline info
	RENDER_INPUT_SCENE renderscene;
	//RENDER_OUTPUT scene_depthtexture;
	std::list <RENDER_OUTPUT> shadow_depthtexturelist;
	RENDER_OUTPUT final;
	RENDER_OUTPUT edgecontrastenhancement_depths;
	RENDER_OUTPUT full_scene_buffer;
	RENDER_OUTPUT bloom_buffer;
	RENDER_OUTPUT blur_buffer;
	
	float camfov;
	float view_distance;
	MATHVECTOR <float, 3> campos;
	QUATERNION <float> camorient;
	
	unsigned int fsaa;
	int lighting; ///<lighting quality; see data/settings/options.config for definition of values
	bool bloom;
	
	void ChangeDisplay(const int width, const int height, const int bpp, const int dbpp, const bool fullscreen, 
			   unsigned int antialiasing, std::ostream & info_output, std::ostream & error_output);
	void SetActiveShader(const std::string name);
	bool LoadShader(const std::string & shaderpath, const std::string & name, std::ostream & info_output, std::ostream & error_output, std::string variant="", std::string variant_defines="");
	void EnableShaders(const std::string & shaderpath, std::ostream & info_output, std::ostream & error_output);
	void DisableShaders();
	void DrawBox(const MATHVECTOR <float, 3> & corner1, const MATHVECTOR <float, 3> & corner2) const;
	void SetupCamera();
	
	void Render(RENDER_INPUT * input, RENDER_OUTPUT & output, std::ostream & error_output);
public:
	GRAPHICS_SDLGL() : surface(NULL),initialized(false),using_shaders(false),max_anisotropy(0),shadows(false),
		       	reflection_status(REFLECTION_DISABLED),drawlist_map1_is_front(true),camfov(45),
			view_distance(10000),fsaa(1),lighting(0),bloom(false)
			{activeshader = shadermap.end();}
	~GRAPHICS_SDLGL() {}
	
	///reflection_type is 0 (low=OFF), 1 (medium=static), 2 (high=dynamic)
	void Init(const std::string shaderpath, const std::string & windowcaption,
		  unsigned int resx, unsigned int resy, unsigned int bpp,
    		  unsigned int depthbpp, bool fullscreen, bool shaders,
		  unsigned int antialiasing, bool enableshadows,
    		  int shadow_distance, int shadow_quality,
   		  int reflection_type,
    		  const std::string & static_reflectionmap_file,
		  int anisotropy, const std::string & texturesize,
    		  int lighting_quality, bool newbloom,
		  std::ostream & info_output, std::ostream & error_output);
	void Deinit();
	void BeginScene(std::ostream & error_output);
	std::map < DRAWABLE_FILTER *, std::vector <SCENEDRAW> > & GetUnusedDrawlistmap()
	{
		if (drawlist_map1_is_front)
			return drawlist_map2;
		else
			return drawlist_map1;
	}
	void SetupScene(float fov, float new_view_distance, const MATHVECTOR <float, 3> cam_position, const QUATERNION <float> & cam_rotation)
	{
		camfov = fov;
		campos = cam_position;
		camorient = cam_rotation;
		view_distance = new_view_distance;
		
		drawlist_map1_is_front = !drawlist_map1_is_front;
	}
	void DrawScene(std::ostream & error_output);
	void EndScene(std::ostream & error_output);
	void Screenshot(std::string filename);
	int GetW() const {return w;}
	int GetH() const {return h;}
	float GetWHRatio() const {return (float)w/h;}
	int GetMaxAnisotropy() const {return max_anisotropy;}
	bool AntialiasingSupported() {return GLEW_ARB_multisample;}

	bool GetUsingShaders() const
	{
		return using_shaders;
	}
	
	bool ReloadShaders(const std::string & shaderpath, std::ostream & info_output, std::ostream & error_output);
};

#endif
