#ifndef _REPLAY_H
#define _REPLAY_H

#include "car.h"
#include "carinput.h"
#include "serialization.h"

#include <iostream>
#include <string>

class REPLAY
{
friend class SERIALIZER;
private:
	//sub-classes
	class REPLAYVERSIONING
	{
		public:
			std::string format_version;
			int inputs_supported;
			float framerate;
			
			REPLAYVERSIONING() : format_version("VDRIFTREPLAYV??"), inputs_supported(0), framerate(0) {}
			REPLAYVERSIONING(const std::string & ver, unsigned int ins, float newfr) : format_version(ver),
				     inputs_supported(ins), framerate(newfr) {}
				     
			void Serialize(SERIALIZER * s)
			{
				_SERIALIZE_(s,inputs_supported);
				_SERIALIZE_(s,framerate);
			}

			void Save(std::ostream & outstream) const
			{
				//write the file format version data manually.  if the serialization functions were used, a variable length string would be written instead, which isns't exactly what we want
				outstream.write(format_version.data(), format_version.length());
				
				//write the rest of the versioning info
				SERIALIZER_BIN_OUTPUT serialize_output(outstream);
				serialize_output.SetPointerTracking(false);
				serialize_output.StartSerialize("versioning", const_cast<REPLAYVERSIONING&>(*this));
			}
			
			void Load(std::istream & instream)
			{
				//read the file format version data manually
				const unsigned int bufsize = format_version.length()+1;
				char version_buf[bufsize];
				instream.get(version_buf, bufsize);
				version_buf[bufsize] = '\0';
				format_version = version_buf;
				
				//read the rest of the versioning info
				SERIALIZER_BIN_INPUT serialize_input(instream);
				serialize_input.SetPointerTracking(false);
				serialize_input.StartSerialize("versioning", *this);
			}
			
			bool operator==(const REPLAYVERSIONING & other) const
			{
				return (format_version == other.format_version && 
					inputs_supported == other.inputs_supported && 
				        framerate == other.framerate);
			}
	};
	
	class INPUTFRAME
	{
	private:
		friend class SERIALIZER;
		class INPUTPAIR
		{
		public:
			INPUTPAIR() : index(0), value(0) {}
			INPUTPAIR(int newindex, float newvalue) : index(newindex), value(newvalue) {}
			
			int index;
			float value;
			
			void Serialize(SERIALIZER * s)
			{
				_SERIALIZE_(s,index);
				_SERIALIZE_(s,value);
			}
		};
		
		int frame;
		std::vector <INPUTPAIR> inputs;
		
	public:
		INPUTFRAME() : frame(0) {}
		INPUTFRAME(int newframe) : frame(newframe) {}
		
		void Serialize(SERIALIZER * s)
		{
			_SERIALIZE_(s,frame);
			_SERIALIZEVECTOR_(s,inputs);
		}
		
		void AddInput(int index, float value)
		{
			inputs.push_back(INPUTPAIR(index,value));
		}
		
		unsigned int GetNumInputs() const {return inputs.size();}

		int GetFrame() const
		{
			return frame;
		}
		
		///returns a pair for the <control id, value> of the indexed input
		std::pair <int, float> GetInput(unsigned int index) const {assert(index < inputs.size());return std::pair <int, float> (inputs[index].index, inputs[index].value);}
	};
	
	class STATEFRAME
	{
	private:
		friend class SERIALIZER;
		
		int frame;
		std::string binary_state_data;
		
	public:
		STATEFRAME() : frame(0) {}
		STATEFRAME(int newframe) : frame(newframe) {}
		
		void Serialize(SERIALIZER * s)
		{
			_SERIALIZE_(s,frame);
			_SERIALIZE_(s,binary_state_data);
			//std::cout << binary_state_data.length() << std::endl;
		}

		void SetBinaryStateData (const std::string & value)
		{
			binary_state_data = value;
			//std::cout << binary_state_data.length() << std::endl;
		}

		int GetFrame() const
		{
			return frame;
		}

		std::string GetBinaryStateData() const
		{
			//std::cout << binary_state_data.length() << std::endl;
			return binary_state_data;
		}
	};
	
	//version information
	const REPLAYVERSIONING version_info;
	
	//serialized
	std::string cartype; //car type, used for loading graphics and sound
	std::string carpaint; //car paint id string
	std::string carfile; //entire contents of the car file (e.g. XS.car)
	std::string track;
	int totalframes;
	std::vector <INPUTFRAME> inputframes;
	std::vector <STATEFRAME> stateframes;
	
	//not stored in the replay file
	int frame;
	enum
	{
		IDLE,
  		RECORDING,
    		PLAYING
	} replaymode;
	std::vector <float> inputbuffer;
	std::string replayfilename;
	std::vector <INPUTFRAME>::iterator cur_inputframe;
	std::vector <STATEFRAME>::iterator cur_stateframe;
	
	//functions
	void ProcessPlayInputFrame(const INPUTFRAME & frame);
	void ProcessPlayStateFrame(const STATEFRAME & frame, CAR & car);
	bool Load(std::istream & instream, std::ostream & error_output); ///< returns true on success
	void Save(std::ostream & outstream);
	
public:
	REPLAY(float framerate) : version_info("VDRIFTREPLAYV09", CARINPUT::GAME_ONLY_INPUTS_START_HERE, framerate),totalframes(0),frame(0),replaymode(IDLE) {inputbuffer.resize(CARINPUT::GAME_ONLY_INPUTS_START_HERE, 0);}
	
	//playing
	bool StartPlaying(const std::string & newreplayfilename, std::ostream & error_output); ///< returns true on success
	void StopPlaying();
	bool GetPlaying() const {return (replaymode == PLAYING);} ///< returns true if the replay system is currently playing
	const std::vector <float> & PlayFrame(CAR & car);
	
	//recording
	void StartRecording(const std::string & newreplayfilename, const std::string & newcartype, const std::string & newcarpaint, const std::string & carfilename, const std::string & trackname, ostream & error_log);
	void StopRecording();
	bool GetRecording() const {return (replaymode == RECORDING);} ///< returns true if the replay system is currently recording
	void RecordFrame(const std::vector <float> & inputs, CAR & car);
	
	void Serialize(SERIALIZER * s)
	{
		_SERIALIZE_(s, cartype);
		_SERIALIZE_(s, carpaint);
		_SERIALIZE_(s, carfile);
		_SERIALIZE_(s, track);
		_SERIALIZE_(s, totalframes);
		_SERIALIZEVECTOR_(s, inputframes);
		_SERIALIZEVECTOR_(s, stateframes);
	}

	std::string GetCarType() const
	{
		return cartype;
	}

	std::string GetCarFile() const
	{
		return carfile;
	}

	std::string GetTrack() const
	{
		return track;
	}

	std::string GetCarPaint() const
	{
		return carpaint;
	}
	
};

#endif
