#ifndef _TIMER_H
#define _TIMER_H

#include "scenegraph.h"
#include "texture.h"
#include "vertexarray.h"
#include "configfile.h"
#include "font.h"

#include <ostream>
#include <string>
#include <vector>
#include <map>

class TIMER
{
private:

	class LAPTIME
	{
		private:
			bool havetime;
			double time;

		public:
			LAPTIME() {Reset();}
			void Reset()
			{
				havetime = false;
				time = 0;
			}
			bool HaveTime() const {return havetime;}
			double GetTimeInSeconds() const {return time;}
			///convert time in seconds into output min and secs
			void GetTimeInMinutesSeconds(float & secs, int & min) const
			{
				min = (int) time / 60;
				secs = time - min*60;
			}
			void Set(double newtime)
			{
				time = newtime;
				havetime = true;
			}
			///only set the time if we don't have a time or if the new time is faster than the current time
			void SetIfFaster(double newtime)
			{
				if (!havetime || newtime < time)
					time = newtime;
				havetime = true;
			}
	};

	class LAPINFO
	{
		private:
			double time; //running time for this lap
			LAPTIME lastlap; //last lap time for player & opponents
			LAPTIME bestlap; //best lap time for player & opponents
			double totaltime; //total time of a race for player & opponents
			int num_laps; //current lap
			std::string cartype;
			double lapdistance; //total track distance driven this lap in meters

		public:
			LAPINFO(const std::string & newcartype) : cartype(newcartype) {Reset();}

			void Reset()
			{
				time = totaltime = 0.0;
				lastlap.Reset();
				bestlap.Reset();
				num_laps = 0;
			}

			void Tick(float dt)
			{
				time += dt;
			}

			void Lap(bool countit)
			{
				if (countit)
				{
					lastlap.Set(time);
					bestlap.SetIfFaster(time);
				}

				totaltime += time;
				time = 0.0;
				num_laps++;
			}

			const std::string & GetCarType() const {return cartype;}

			void DebugPrint(std::ostream & out)
			{
				out << "car=" << cartype << ", t=" << totaltime << ", tlap=" << time << ", last=" <<
					lastlap.GetTimeInSeconds() << ", best=" << bestlap.GetTimeInSeconds() <<
					", lap=" << num_laps << std::endl;
			}

			double GetTime() const
			{
				return time;
			}

			double GetLastLap() const
			{
				return lastlap.GetTimeInSeconds();
			}

			double GetBestLap() const
			{
				return bestlap.GetTimeInSeconds();
			}

			int GetCurrentLap() const
			{
			    return num_laps;
			}

			void UpdateLapDistance(double newdistance)
			{
			    lapdistance = newdistance;
			}

			double GetLapDistance() const {return lapdistance;}
	};

	std::vector <LAPINFO> car;

	//variables for drawing text

	bool loaded;
	CONFIGFILE trackrecords; //the track records configfile
	std::string trackrecordsfile; //the filename for the track records
	float pretime; //amount of time left in staging
	unsigned int playercarindex; //the index for the player's car; defaults to zero

public:
	TIMER() : loaded(false),pretime(0.0),playercarindex(0) {}
	~TIMER() {Unload();}

	bool Load(const std::string & trackrecordspath, float stagingtime, std::ostream & error_output);
	///add a car of the given type and return the integer identifier that the track system will use
	int AddCar(const std::string & cartype) {car.push_back(LAPINFO(cartype));return car.size()-1;}
	void SetPlayerCarID(int newid) {playercarindex = newid;}
	void Unload();
	bool Staging() const {return (pretime > 0);}
	void Tick(float dt);
	void Lap(const unsigned int carid, const int prevsector, const int nextsector, const bool countit);
	void UpdateDistance(const unsigned int carid, const double newdistance)
	{
	    assert(carid < car.size());
	    car[carid].UpdateLapDistance(newdistance);
	}
	void DebugPrint(std::ostream & out)
	{
		for (unsigned int i = 0; i < car.size(); i++)
		{
			out << i << ". ";
			car[i].DebugPrint(out);
		}
	}
	float GetPlayerTime() {assert(playercarindex<car.size());return car[playercarindex].GetTime();}
	float GetLastLap() {assert(playercarindex<car.size());return car[playercarindex].GetLastLap();}
	float GetBestLap()
	{
		assert(playercarindex<car.size());
		float curbestlap = car[playercarindex].GetBestLap();
		float prevbest(0);
		bool haveprevbest = trackrecords.GetParam(car[playercarindex].GetCarType() + ".sector 0", prevbest);
		if (haveprevbest)
		{
			if (curbestlap == 0)
				return prevbest;
			else
				if (prevbest < curbestlap)
					return prevbest;
				else
					return curbestlap;
		}
		else
			return curbestlap;
	}
	int GetPlayerCurrentLap() {return GetCurrentLap(playercarindex);}
	int GetCurrentLap(unsigned int index) {assert(index<car.size());return car[index].GetCurrentLap();}
	float GetStagingTimeLeft() const {return pretime;}

	///return the place (first element) out of total (second element)
	std::pair <int, int> GetCarPlace(int index);

	std::pair <int, int> GetPlayerPlace() {return GetCarPlace(playercarindex);}
};

#endif
